//############################################################################//
#define ORBITER_MODULE
#define _CRT_SECURE_NO_WARNINGS
#define _CRT_NONSTDC_NO_DEPRECATE
#include <math.h>
#include <stdio.h>
#include "orbitersdk.h"
//############################################################################//
#include "Arm.h"
//############################################################################//
//Class creation
Arm::Arm(OBJHANDLE hObj,int fmodel):VESSEL2(hObj,fmodel){}
Arm::~Arm(){}
//############################################################################//
//Fill animation definition
void Arm::make_anim(int n,int id,int tp,int repeat,int trigkey,int status,double spd,double proc,double init,int grpcnt,int compcnt)
{
 ani[n].id=id;
 ani[n].tp=tp;
 ani[n].repeat=repeat;
 ani[n].trigkey=trigkey;
 ani[n].status=status;
 ani[n].spd=spd;
 ani[n].proc=proc;
 ani[n].init=init;
 ani[n].grpcnt=grpcnt;
 ani[n].compcnt=compcnt;
}
//############################################################################//
//Fill animation component definition
void Arm::make_anim_comp(int n,int c,int msh,int id,int tp,int grp,int parent,VECTOR3 ref,VECTOR3 axshsc,double angle,double st0,double st1)
{
 ani[n].comp[c].msh=msh;
 ani[n].comp[c].id=id;
 ani[n].comp[c].tp=tp;
 ani[n].comp[c].grp=grp;
 ani[n].comp[c].parent=parent;
 ani[n].comp[c].ref=ref;
 ani[n].comp[c].axshsc=axshsc;
 ani[n].comp[c].angle=(float)angle;
 ani[n].comp[c].st0=st0;
 ani[n].comp[c].st1=st1;
}
//############################################################################//
//Fill attachment point definition
void Arm::make_attachment(int n,char *name,char *id,bool loose,bool ischild,double range,VECTOR3 pos,VECTOR3 dir,VECTOR3 rot)
{
 strcpy(att[n].id,id);
 strcpy(att[n].name,name);
 att[n].ischild=ischild;
 att[n].loose=loose;
 att[n].range=range;
 att[n].pos=pos;
 att[n].dir=dir;
 att[n].rot=rot;
}
//############################################################################//
//Define animations
void Arm::define_animations()
{
 int i,j,k,n,a,b;
 //Sequence 0
 make_anim(0,0,0,0,0,-2,0.083,0.5,0.5,1,1);

 ani[0].grp=(UINT**)malloc(1*sizeof(UINT*));
 static UINT g0[1+1]={1,5}; ani[0].grp[0]=g0;

 ani[0].comp=(animcompinfo*)malloc(1*sizeof(animcompinfo));
 make_anim_comp(0,0,0,0,0,0,0,_V(0,0,0),_V(0,1,0),2.094,0,1);

 //Sequence 1
 make_anim(1,0,0,0,0,-2,0.083,0,0,1,1);

 ani[1].grp=(UINT**)malloc(1*sizeof(UINT*));
 static UINT g1[1+1]={1,4}; ani[1].grp[0]=g1;

 ani[1].comp=(animcompinfo*)malloc(1*sizeof(animcompinfo));
 make_anim_comp(1,0,0,1,0,0,0,_V(0,0,23.704),_V(0,1,0),6.283,0,1);

 //Sequence 2
 make_anim(2,0,0,0,0,-2,0.025,0,0,1,1);

 ani[2].grp=(UINT**)malloc(1*sizeof(UINT*));
 static UINT g2[1+1]={1,3}; ani[2].grp[0]=g2;

 ani[2].comp=(animcompinfo*)malloc(1*sizeof(animcompinfo));
 make_anim_comp(2,0,0,2,0,0,1,_V(0,0,0),_V(0,1,0),4.887,0,1);

 //Sequence 3
 make_anim(3,0,0,0,0,-2,0.029,0.5,0.5,1,1);

 ani[3].grp=(UINT**)malloc(1*sizeof(UINT*));
 static UINT g3[1+1]={1,2}; ani[3].grp[0]=g3;

 ani[3].comp=(animcompinfo*)malloc(1*sizeof(animcompinfo));
 make_anim_comp(3,0,0,3,0,0,2,_V(0,0,23.704),_V(0,1,0),-5.760,0,1);

 //Sequence 4
 make_anim(4,0,0,0,0,-2,0.083,0.5,0.5,1,1);

 ani[4].grp=(UINT**)malloc(1*sizeof(UINT*));
 static UINT g4[1+1]={1,1}; ani[4].grp[0]=g4;

 ani[4].comp=(animcompinfo*)malloc(1*sizeof(animcompinfo));
 make_anim_comp(4,0,0,4,0,0,3,_V(0,0,0),_V(0,1,0),5.760,0,1);

 //Sequence 5
 make_anim(5,0,0,0,0,-2,0.083,1,1,1,1);

 ani[5].grp=(UINT**)malloc(1*sizeof(UINT*));
 static UINT g5[1+1]={1,0}; ani[5].grp[0]=g5;

 ani[5].comp=(animcompinfo*)malloc(1*sizeof(animcompinfo));
 make_anim_comp(5,0,0,5,0,0,4,_V(0,0,23.704),_V(0,1,0),4.014,0,1);

 //Sequence 6
 make_anim(6,0,0,0,0,-2,0.032,1,1,1,1);

 ani[6].grp=(UINT**)malloc(1*sizeof(UINT*));
 static UINT g6[0+1]={0}; ani[6].grp[0]=g6;

 ani[6].comp=(animcompinfo*)malloc(1*sizeof(animcompinfo));
 make_anim_comp(6,0,0,6,0,0,5,_V(0,0,0),_V(0,1,0),6.283,0,1);

 //Zero out components
 for(i=0;i<ANIM_COUNT;i++)for(j=0;j<ani[i].compcnt;j++)ani[i].comp[j].cmpt=NULL;
 
 //Animation core creation
 n=0;
 for(i=0;i<ANIM_COUNT;i++){
  for(j=0;j<ani[i].compcnt;j++){
   if(ani[i].comp[j].id==ARM_GRAPPLE_SEQUENCE){
    ani[i].comp[j].cmpdef=new MGROUP_ROTATE(LOCALVERTEXLIST,MAKEGROUPARRAY(arm_tip),3,ani[i].comp[j].ref,ani[i].comp[j].axshsc,ani[i].comp[j].angle);
   }else{
    if(ani[i].comp[j].tp==0)ani[i].comp[j].cmpdef=new MGROUP_ROTATE   (ani[i].comp[j].msh,&(ani[i].grp[ani[i].comp[j].grp][1]),ani[i].grp[ani[i].comp[j].grp][0],ani[i].comp[j].ref,ani[i].comp[j].axshsc,ani[i].comp[j].angle);
    if(ani[i].comp[j].tp==1)ani[i].comp[j].cmpdef=new MGROUP_TRANSLATE(ani[i].comp[j].msh,&(ani[i].grp[ani[i].comp[j].grp][1]),ani[i].grp[ani[i].comp[j].grp][0],ani[i].comp[j].axshsc);
    if(ani[i].comp[j].tp==2)ani[i].comp[j].cmpdef=new MGROUP_SCALE    (ani[i].comp[j].msh,&(ani[i].grp[ani[i].comp[j].grp][1]),ani[i].grp[ani[i].comp[j].grp][0],ani[i].comp[j].ref,ani[i].comp[j].axshsc);
   }
  }
  ani[i].id=CreateAnimation(ani[i].init);
  n+=ani[i].compcnt;
 }

 //Animation sequences
 //SC3 have a pequliar method of arranging sequences, so we have to sort them accordingly
 //FIXME: Lousy implementation, move to code generator?
 k=0;
 for(k=0;k<n;k++){
  for(i=0;i<ANIM_COUNT;i++){
   for(j=0;j<ani[i].compcnt;j++){
    if(ani[i].comp[j].id!=k)continue;
    int p=ani[i].comp[j].parent;
    if((p>=0)&(p<n)){
     ANIMATIONCOMPONENT_HANDLE prnt=NULL;
     for(a=0;a<ANIM_COUNT;a++)for(b=0;b<ani[a].compcnt;b++)if(ani[a].comp[b].id==p)prnt=ani[a].comp[b].cmpt;
     ani[i].comp[j].cmpt=AddAnimationComponent(ani[i].id,ani[i].comp[j].st0,ani[i].comp[j].st1,ani[i].comp[j].cmpdef,prnt);
    }else ani[i].comp[j].cmpt=AddAnimationComponent(ani[i].id,ani[i].comp[j].st0,ani[i].comp[j].st1,ani[i].comp[j].cmpdef);
    break;
   }
  }
 }
}
//############################################################################//
//All the parameters
void Arm::clbkSetClassCaps(FILEHANDLE cfg)
{
 int i;

 //Base stuff
 SetSize(190);
 SetEmptyMass(100);
 SetPMI(_V(18,18,0.1));
 SetCrossSections(_V(4.7,4.4,0.4));
 SetRotDrag(_V(0,0,0));
 SetTouchdownPoints(_V(0,0,1),_V(-1,0,-1),_V(1,0,-1));
 SetSurfaceFrictionCoeff(0.07,0.3);
 SetCW(0.09,0.09,2,1.4);
 EnableTransponder(true);
 InitNavRadios(4);
 SetCameraOffset(_V(-0.4,0.5,-5.45));

 SetVisibilityLimit(0,0.002);
 SetAlbedoRGB(_V((float)255/255,(float)255/255,(float)255/255));
 SetGravityGradientDamping(0);
 SetWingAspect(0);
 SetWingEffectiveness(0);
 SetMaxWheelbrakeForce(200000);
 SetBankMomentScale(0);
 SetPitchMomentScale(0);

 //Attachments
 att_on=0;
 att_sel=0;
 att_blank=0;
 make_attachment(0,"Rms","",false,false,1,_V(-0.2,-0.21,5),_V(1,0,0),_V(0,0,1));
 make_attachment(1,"Attach","XS",false,true,0,_V(0,-7.519,0),_V(0,-1,0),_V(1,0,0));
 for(i=0;i<ATTACH_COUNT;i++)att_h[i]=CreateAttachment(att[i].ischild,att[i].pos,att[i].dir,att[i].rot,att[i].id,att[i].loose);
 arm_on=0;
 arm_sel=0;
 arm_blank=0;
 arm_tip[0]=_V(0,0,0);
 arm_tip[1]=_V(0,1,0);
 arm_tip[2]=_V(1,0,0);
 strcpy(arm[0].name,"yaw1");
 arm[0].seq=0;
 arm[0].range_low=0;
 arm[0].range_high=360;
 strcpy(arm[1].name,"yaw2");
 arm[1].seq=1;
 arm[1].range_low=-180;
 arm[1].range_high=180;
 strcpy(arm[2].name,"yaw3");
 arm[2].seq=2;
 arm[2].range_low=-180;
 arm[2].range_high=180;
 strcpy(arm[3].name,"yaw4");
 arm[3].seq=3;
 arm[3].range_low=-180;
 arm[3].range_high=180;
 strcpy(arm[4].name,"yaw5");
 arm[4].seq=4;
 arm[4].range_low=-180;
 arm[4].range_high=180;
 strcpy(arm[5].name,"yaw6");
 arm[5].seq=5;
 arm[5].range_low=-180;
 arm[5].range_high=180;
 strcpy(arm[6].name,"Rotate Grapple point");
 arm[6].seq=6;
 arm[6].range_low=-180;
 arm[6].range_high=180;
 VECTOR3 xp,xr;
 xp=arm_tip[1]-arm_tip[0];normalise(xp);
 xr=arm_tip[2]-arm_tip[0];normalise(xr);
 SetAttachmentParams(att_h[ARM_GRAPPLE_ATTACH],arm_tip[0],xp,xr);

 //Meshes
 SetMeshVisibilityMode(msh_idh[0]=AddMesh(msh_h[0]=oapiLoadMeshGlobal("pegasus\\Arm")),3);

 //Animations
 define_animations();
}
//############################################################################//
void Arm::clbkLoadStateEx(FILEHANDLE scn,void *vs)
{
 char *line;
 int i,n;

 while(oapiReadScenario_nextline(scn,line)){
  //Animation state loading
  if(!strnicmp(line,"ANIM_",5)){
   sscanf(line+5,"%d",&i);
   if((i>=0)&&(i<ANIM_COUNT)){
    n=0;
    if(i>=10)n=1;
    if(i>=100)n=2;
    if(i>=1000)n=3;
    sscanf(line+6+n,"%d",&ani[i].status);
    sscanf(line+6+n+3,"%lf",&ani[i].proc);
    SetAnimation(ani[i].id,ani[i].proc);
   }
  }else{
   ParseScenarioLineEx(line,vs);
  }
 }
}
//############################################################################//
void Arm::clbkSaveState(FILEHANDLE scn)
{
 SaveDefaultState(scn);
 
 //Animation state
 char cbuf[256],nbuf[256];
 int i;

 for(i=0;i<ANIM_COUNT;i++){
  sprintf(cbuf,"%d %0.4f",ani[i].status,ani[i].proc);
  sprintf(nbuf,"ANIM_%d",i);
  oapiWriteScenario_string(scn,nbuf,cbuf);
 }
}
//############################################################################//
 void Arm::toggle_attach(int point)
{
 VESSEL *v;
 DWORD nAttach,i,j;
 double dst,min_dst;
 VECTOR3 gpos,grms,pos,dir,rot;
 OBJHANDLE hv,min_hv;
 ATTACHMENTHANDLE tgt_att,min_att;
 //const char *id;
 
 min_dst=1e50;
 min_att=NULL;
 min_hv=NULL;
 
 hv=GetAttachmentStatus(att_h[point]);
 
 if(hv){
  //release
  DetachChild(att_h[point]);
 }else{
  //grapple
  Local2Global(att[point].pos,grms);  //global position of RMS tip
 
  for(i=0;i<oapiGetVesselCount();i++){
   hv=oapiGetVesselByIndex(i);
   if(hv==GetHandle())continue;
   oapiGetGlobalPos(hv,&gpos);
   dst=dist(gpos,grms);
   if(dst<oapiGetSize(hv)+oapiGetSize(GetHandle())){
    v=oapiGetVesselInterface(hv);
    nAttach=v->AttachmentCount(true);
    for(j=0;j<nAttach;j++){
     tgt_att=v->GetAttachmentHandle(true,j);
     if(v->GetAttachmentStatus(tgt_att))continue;
     //id=v->GetAttachmentId(tgt_att);
     //if(strncmp(id,"GS",2))continue;
     v->GetAttachmentParams(tgt_att,pos,dir,rot);
     v->Local2Global(pos,gpos);
     dst=dist(gpos,grms);
					if((dst<att[point].range)&&(dst<min_dst)){
						min_dst=dst;
      min_att=tgt_att;
      min_hv=hv;
					}
    }
   }
  }
  //Attach closest
  if(min_att)AttachChild(min_hv,att_h[point],min_att);
 }
}
//############################################################################//
int  Arm::clbkConsumeBufferedKey(DWORD key,bool down,char *kstate)
{
 int i;
 if(!down)return 0;
 if(KEYMOD_SHIFT(kstate)){ 
  if(arm_on)switch(key){
   case OAPI_KEY_NUMPAD4:arm_sel--;if(arm_sel<0)arm_sel=ARM_COUNT-1;break;
   case OAPI_KEY_NUMPAD6:arm_sel++;if(arm_sel>=ARM_COUNT)arm_sel=0;break;
   case OAPI_KEY_NUMPAD0:toggle_attach(ARM_GRAPPLE_ATTACH);break;
  }
 }else switch(key){
  case OAPI_KEY_SPACE:arm_on=!arm_on;arm_blank=1;break;
 }
 if(KEYMOD_SHIFT(kstate)){ 
  if(att_on)switch(key){
   case OAPI_KEY_NUMPAD4:att_sel--;if(att_sel<0)att_sel=ATTACH_COUNT-1;break;
   case OAPI_KEY_NUMPAD6:att_sel++;if(att_sel>=ATTACH_COUNT)att_sel=0;break;
   case OAPI_KEY_NUMPAD0:toggle_attach(att_sel);break;
  }
 }else switch(key){
  case OAPI_KEY_A:att_on=!att_on;att_blank=1;break;
 }
 for(i=0;i<ANIM_COUNT;i++)if(ani[i].tp==0)if(key==ani[i].trigkey){
  //Animation input
  if(!ani[i].repeat){
   switch(ani[i].status){
    //Fix it short
    case -2:ani[i].status=1;break;
    case -1:ani[i].status=1;break;
    case  1:ani[i].status=-1;break;
    case  2:ani[i].status=-1;break;
    default:ani[i].status=2;
   }
  }else{
   switch(ani[i].status){
    case 1:ani[i].status=0;break;
    case 0:ani[i].status=1;break;
    default:ani[i].status=0;
   }
  }
 }
 return 0;
}
//#######################################################################################//
int Arm::clbkConsumeDirectKey(char *kstate)
{     
 int i;
 for(i=0;i<ARM_COUNT;i++)ani[arm[i].seq].status=0;
 if(arm_on){
  if(KEYMOD_SHIFT(kstate)){
   if(KEYDOWN(kstate,OAPI_KEY_NUMPAD2)){
    ani[arm[arm_sel].seq].status=1;
    RESETKEY(kstate,OAPI_KEY_NUMPAD8);
   }
   if(KEYDOWN(kstate,OAPI_KEY_NUMPAD8)){
    ani[arm[arm_sel].seq].status=-1;
    RESETKEY(kstate,OAPI_KEY_NUMPAD8);
   }
  }  
 }
 return 0;
}
//#######################################################################################//
void Arm::clbkPostStep(double simt,double simdt,double mjd)
{
 double pos;
 //Animation timing controls
 int i;
 for(i=0;i<ANIM_COUNT;i++)if(abs(ani[i].status)==1){
  double da=simdt*ani[i].spd;
  if(ani[i].repeat){
   ani[i].proc=fmod(ani[i].proc+da,1);
  }else{
   if(ani[i].status==-1){
    if(ani[i].proc>0.0)ani[i].proc=max(0.0,ani[i].proc-da);
    else                     ani[i].status=-2;
   }else{
    if(ani[i].proc<1.0)ani[i].proc=min(1.0,ani[i].proc+da);
    else                     ani[i].status=2;
   }
  }
  SetAnimation(ani[i].id,ani[i].proc);
 }
 if(arm_on)if((arm_sel>=0)&&(arm_sel<ARM_COUNT)){
  pos=ani[arm[arm_sel].seq].proc;
  pos=arm[arm_sel].range_low+pos*(arm[arm_sel].range_high-arm[arm_sel].range_low);
  VECTOR3 xp,xr;
  xp=arm_tip[1]-arm_tip[0];normalise(xp);
  xr=arm_tip[2]-arm_tip[0];normalise(xr);
  SetAttachmentParams(att_h[ARM_GRAPPLE_ATTACH],arm_tip[0],xp,xr);
  att[ARM_GRAPPLE_ATTACH].pos=arm_tip[0];
  att[ARM_GRAPPLE_ATTACH].dir=xp;
  att[ARM_GRAPPLE_ATTACH].rot=xr;
  sprintf(oapiDebugString(),"Robotic Arm: %s (%.3f)",arm[arm_sel].name,pos);
 }
 if(arm_blank){sprintf(oapiDebugString(),"");arm_blank=0;}
 if(att_on)if((att_sel>=0)&&(att_sel<ATTACH_COUNT)){
  sprintf(oapiDebugString(),"Attach:%s:%s",att[att_sel].name,att[att_sel].name);
 }
 if(att_blank){sprintf(oapiDebugString(),"");att_blank=0;}
}
//#######################################################################################//
//Creation
DLLCLBK void InitModule(HINSTANCE hModule){}
DLLCLBK void ExitModule(HINSTANCE hModule){}
DLLCLBK VESSEL *ovcInit(OBJHANDLE hvessel,int flightmodel){return new Arm(hvessel,flightmodel);}
DLLCLBK void ovcExit(VESSEL *vessel){if(vessel)delete(Arm*)vessel;}
//#######################################################################################//
